
/* 

   Sherman's aquarium - Screensaver part

   Updated and partly rewritten for Sherman's aquarium v3.0.0 on
   30th and 31st December 2003.

   Jonas Aaberg <cja@gmx.net>

*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <SDL/SDL.h>
#include <X11/Xlib.h>



#if SDL_BYTEORDER == SDL_BIG_ENDIAN
#define RMASK 0xff000000
#define GMASK 0x00ff0000
#define BMASK 0x0000ff00
#define AMASK 0x000000ff

#else

#define RMASK 0x000000ff
#define GMASK 0x0000ff00
#define BMASK 0x00ff0000
#define AMASK 0xff000000
#endif


/* Stolen from gai.h :-) */
typedef struct
{
    char *name;
    char *version;
    char *nice_name;
    char *author;
    char *license;
    char *description;
    char *icon;
    char *image_path;
} GaiApplet;


#include "../config.h"
#include "aquarium.h"
#include "bottom.h"
#include "bubble.h"
#include "background.h"
#include "fish.h"
#include "soundeffects.h"
#include "vroot.h"

#ifdef DARWIN
#include "getopt.h"
#else
#include <getopt.h>
#endif


#define ARG_SEAFLOOR 1
#define ARG_SOUND 2
#define ARG_PLANTS 3
#define ARG_PLANTSCALE 4
#define ARG_BOTTOMANIMALS 5
#define ARG_OGG 6
#define ARG_BG_SOLID 7
#define ARG_BG_SHADED 8
#define ARG_BG_WATERALIKE 9
#define ARG_BG_IMAGE 10
#define ARG_LCR 11
#define ARG_LCG 12
#define ARG_LCB 13
#define ARG_UCR 14
#define ARG_UCG 15
#define ARG_UCB 16
#define ARG_BG_IMAGE_FILE 17
#define ARG_RANDOM 18
#define ARG_SELECTED 19
#define ARG_RANDOM_POP 20
#define ARG_FISH1 21
#define ARG_FISH2 22
#define ARG_FISH3 23
#define ARG_FISH4 24
#define ARG_FISH5 25
#define ARG_FISH6 26
#define ARG_BLOWFISH 27
#define ARG_SWORDFISH 28
#define ARG_BDWELLER 29
#define ARG_FILLMORE 30
#define ARG_SHERMAN 31
#define ARG_MEGAN 32
#define ARG_HUNTER 33
#define ARG_PREY 34
#define ARG_LORI 35
#define ARG_ERNEST 36
#define ARG_SQUID 37
#define ARG_HAWTHORNE 38
#define ARG_EAT 39
#define ARG_EXPLODE 40
#define ARG_REBIRTH 41
#define ARG_SCALEDIFF 42
#define ARG_SPEED 43
#define ARG_SCALE 44
#define ARG_HUNTERA 45
#define ARG_SWORDA 46
#define ARG_SOUNDPRG 47
#define ARG_WINDOW_ID 48
#define ARG_FULLSCREEN 49
#define ARG_BUBBLE 50




#define DEPTH 24

static int screen_width;
static int screen_height;



static SDL_Surface *screen=NULL, *screen_image, *background, **thisfish;
static SDL_Rect *fish_dest, *fish_src, *clean_dest;
static int curr_dest, clean_count,  no_sdl_quit = 0;
int window_id = -1, fullscreen = 0;

static AquariumData ad;


void screensaver_draw_image(int x, int y, int idx, int rev, SA_Image *image)
{

    fish_dest[curr_dest].x=x;
    fish_dest[curr_dest].y=y;
    fish_dest[curr_dest].w=image->width;
    fish_dest[curr_dest].h=image->height;

    fish_src[curr_dest].x=0;
    fish_src[curr_dest].y=(int)((float)idx*(float)image->full_height/(float)image->frames+0.5);
    fish_src[curr_dest].w=image->width;
    fish_src[curr_dest].h=image->height;

    if(!rev)
	thisfish[curr_dest] = SDL_CreateRGBSurfaceFrom(image->image,
						       image->width,
						       (int)((float)image->full_height*(float)(idx+1) / 
							     (float)image->frames + 0.5),
						       32,image->rowstride,
						       RMASK, GMASK, BMASK, AMASK);
    else
	thisfish[curr_dest] = SDL_CreateRGBSurfaceFrom(image->rev,
						       image->width,
						       (int)((float)image->full_height*(float)(idx+1) / 
							     (float)image->frames + 0.5),
						       32,image->rowstride,
						       RMASK, GMASK, BMASK, AMASK);

    curr_dest++;
}




void screensaver_clean(int x,int y,int w,int h)
{

    clean_dest[clean_count].x=x-5;
    clean_dest[clean_count].y=y-5;
    clean_dest[clean_count].w=w+10;
    clean_dest[clean_count].h=h+10;


    SDL_BlitSurface(background,&clean_dest[clean_count],screen,&clean_dest[clean_count]);

    clean_count++;
}


void screensaver_quit()
{
    /* Resetting the term signal to the orignal so we can quit nicely.*/
    signal(SIGTERM, SIG_DFL);

    if(no_sdl_quit)
	kill(getpid(),SIGTERM);
    else
	exit(0);

    /* In case something is really weird */
    kill(getpid(),SIGKILL);
}


void screensaver_main(void)
{
    SDL_Event event;
    clock_t totaltime1, totaltime2;
    clock_t cali1, cali2;

    int num_events=0;
    int main_loop=0;
    int totalframes=0, frames=0;
    int i;
    int delay=0;


    totaltime1 = clock();
    cali1 = clock();


    while(!main_loop){

	curr_dest=0;
	clean_count=0;

	fish_update();
	bubble_update();

	for(i=0;i<curr_dest;i++){
	    SDL_BlitSurface(thisfish[i],&fish_src[i],screen,&fish_dest[i]);
	}

	/* If we get a SIGTERM from screensaver in this loop, and we later do a SDL_Quit()
	   X will get problems. So we have to avoid calling SDL_Quit if a sigterm is caught
	   in this loop.

	   Otherwise we get:
	   Xlib: unexpected async reply (sequence 0xf03)!
	   And sherman's starts eating processor power like hell!
	*/
	no_sdl_quit = 1;
	for(i=0;i<curr_dest;i++){
	    SDL_UpdateRects(screen,1,&clean_dest[i]);
	    SDL_FreeSurface(thisfish[i]);
	}
	no_sdl_quit = 0;

	if(window_id!=-1)
	    SDL_Flip(screen);

	totalframes++;
	frames++;

	if(frames==10){
	    cali2=clock();

	    /* Check if we're going too fast! */
	    if((float)(cali2-cali1) < (float)(0.2*CLOCKS_PER_SEC)){
		delay=(int)(((float)((0.2*CLOCKS_PER_SEC)-(cali2-cali1))/CLOCKS_PER_SEC)*100000);
	    }
	    else{
		if(delay!=0){
		    if(delay<((int)(((float)((0.2*CLOCKS_PER_SEC)-(cali2-cali1))/CLOCKS_PER_SEC)*100000))){
			delay-=(int)(((float)((0.2*CLOCKS_PER_SEC)-(cali2-cali1))/CLOCKS_PER_SEC)*100000);
		    }
		    else
			delay=0;
		}
	    }
	    /*      printf("delay:%d\n",delay);*/
	    cali1=clock();
	    frames=0;
	}
	usleep(delay);

	while(SDL_PollEvent(&event)){
	    switch(event.type){
	    case SDL_QUIT:
		main_loop=1;
		break;
	    case SDL_KEYDOWN:
		num_events++;
		if(num_events==2)
		    main_loop=1;
		break;
	    case SDL_MOUSEMOTION:
		num_events++;
		if(num_events==2)
		    main_loop=1;
		break;
	    case  SDL_MOUSEBUTTONDOWN:
		num_events++;
		if(num_events==2)
		    main_loop=1;
		break;
	    }
	}    
    }


    totaltime2=clock();
    /*    printf("Frames: %d\n",totalframes);
    printf("Seconds: %f\n",(float)(totaltime2-totaltime1)/CLOCKS_PER_SEC);
    printf("FPS: %f\n",(float)totalframes/((float)(totaltime2-totaltime1)/CLOCKS_PER_SEC));*/

}


void screensaver_init()
{
    char *sdl_command;
    XWindowAttributes win_attr;
    Display *display;
    Fish_settings *fish_settings;
    Bubble_settings *bubble_settings;
    int sdl_flags = SDL_DOUBLEBUF|SDL_HWSURFACE|SDL_ANYFORMAT;

    screen_height = 480;
    screen_width = 640;
    
    if(window_id != -1){
	//	printf("Got window id!\n");
	sdl_command = g_strdup_printf("SDL_WINDOWID=%d",window_id);
	putenv(sdl_command);
	display = XOpenDisplay(NULL);
	XGetWindowAttributes(display, (Window)window_id, &win_attr);
	screen_height = win_attr.height;
	screen_width = win_attr.width;
	fullscreen = 0;
	ad.proximity = 1;	 		/* No sound effects */
    }

    if(fullscreen){
	screen_width = gdk_screen_width();
	screen_height = gdk_screen_height();

	display=XOpenDisplay(NULL);
	sdl_command = g_strdup_printf("SDL_WINDOWID=%d",
				      (int)RootWindowOfScreen(ScreenOfDisplay(display, DefaultScreen(display))));
	putenv(sdl_command);
	ad.proximity = 0;
	sdl_flags |= SDL_FULLSCREEN;
    }

    ad.xmax = ad.windowsize_x = screen_width;
    ad.ymax = ad.windowsize_x = screen_height;
  
    ad.virtual_aquarium_x = ad.xmax + 2 * VIRTUAL_AQUARIUM_DX;
    ad.virtual_aquarium_y = ad.ymax + 2 * VIRTUAL_AQUARIUM_DY;

    ad.ymin = ad.xmin = ad.viewpoint_start_x = ad.viewpoint_start_y = 0;

    ad.rgb = g_malloc0(ad.xmax*3*ad.ymax);
    ad.bgr = g_malloc0(ad.xmax*3*ad.ymax);


    background_init();
    fish_init();
    fish_turn();
    bottom_init();
    bubble_init();

    memcpy(ad.rgb,ad.bgr,ad.xmax*3*ad.ymax);


    if (SDL_Init(SDL_INIT_VIDEO) < 0){
	printf("Can't init SDL: %s\n",SDL_GetError());
	exit(1);
    }


    signal(SIGTERM, screensaver_quit);
    atexit(SDL_Quit);
    

    if(!SDL_VideoModeOK(screen_width,screen_height, DEPTH, sdl_flags)){
	printf("Sorry, video mode %dx%d in %d bits isn't supported by hardware\n",
	       screen_width,screen_height, DEPTH);
	exit(2);
    }


    screen = SDL_SetVideoMode(screen_width, screen_height, DEPTH, sdl_flags);

    if(screen == NULL){
	printf("Unable to set video mode %dx%d in %d bits.\n",
	       screen_width,screen_height,DEPTH);
	exit(3);
    }

    SDL_WM_SetCaption("Sherman's aquarium",NULL);


    /* Hide the mouse cursor */
    if(fullscreen)
	SDL_ShowCursor(0);

    /* Start with all black */
    SDL_FillRect(screen,NULL,0x000000);

    fish_settings = fish_get_settings_ptr();
    bubble_settings = bubble_get_settings_ptr();

    fish_dest = g_malloc0(sizeof(SDL_Rect)*(fish_settings->num_fish + 
					    bubble_settings->max_bubbles));
    fish_src = g_malloc0(sizeof(SDL_Rect)*(fish_settings->num_fish +
					   bubble_settings->max_bubbles));
    clean_dest = g_malloc0(sizeof(SDL_Rect)*(fish_settings->num_fish  + 
					     bubble_settings->max_bubbles));

    thisfish = g_malloc0(sizeof(SDL_Surface*)*(fish_settings->num_fish + bubble_settings->max_bubbles));
    
    
    screen_image = SDL_CreateRGBSurfaceFrom(ad.rgb, ad.xmax, ad.ymax, DEPTH, ad.xmax*3, 
					    RMASK, GMASK, BMASK, 0);
    background = SDL_CreateRGBSurfaceFrom(ad.bgr, ad.xmax, ad.ymax, DEPTH, ad.xmax*3, 
					  RMASK, GMASK, BMASK, 0);


    SDL_BlitSurface(screen_image, NULL, screen, NULL);
    SDL_UpdateRect(screen,0,0,0,0);

}

int htoi(char *str)
{
  int i,sum=0,d;

  if(str[0]!='0' || str[1]!='x') return -1;
  for(i=2;i<strlen(str);i++){
    d=0;
    if(str[i]>='0' && str[i]<='9') d=(int)(str[i]-'0');
    if(str[i]>='A' && str[i]<='F') d=(int)(str[i]-'A'+10);
    if(str[i]>='a' && str[i]<='f') d=(int)(str[i]-'a'+10);

    sum+=d;
    sum=sum<<4;
  }

  return(sum>>4);
}

void screensaver_init_param(int argc, char **argv)
{
    int c, numfish = 0;
    Sound_settings *s, sound_settings = {0,1,0, NULL};
    Bubble_settings *bub, bubble_settings = {20};
    Bottom_settings *b, bottom_settings = {0,5,1,75,2};
    Background_settings *bg, background_settings = {NULL,NULL, 0,1,0,0, 
						    0, 100, 150, 0,
						    10,120, 250, 0,
						    0,0,0,0};

    Fish_settings *f, fish_settings = {0, 0, 75, 0, 100, 0, 15, 0, 0, 0, 75, 75,
				       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1};



    struct option cmdline_options[] = {
	{"sound", no_argument, NULL, ARG_SOUND},
	{"soundprg", required_argument, NULL, ARG_SOUNDPRG},
	{"ogg", no_argument, NULL, ARG_OGG},


	{"seafloor", no_argument, NULL, ARG_SEAFLOOR},
	{"plants", required_argument, NULL, ARG_PLANTS},
	{"plantscale", required_argument, NULL, ARG_PLANTSCALE},
	{"bottomanimals", required_argument, NULL, ARG_BOTTOMANIMALS},

	{"bg_solid", no_argument, NULL, ARG_BG_SOLID},
	{"bg_shaded", no_argument, NULL, ARG_BG_SHADED},
	{"bg_wateralike", no_argument, NULL, ARG_BG_WATERALIKE},
	{"bg_image", no_argument, NULL, ARG_BG_IMAGE},


	{"lcr", required_argument, NULL, ARG_LCR},
	{"lcg", required_argument, NULL, ARG_LCG},
	{"lcb", required_argument, NULL, ARG_LCB},

	{"ucr", required_argument, NULL, ARG_UCR},
	{"ucg", required_argument, NULL, ARG_UCG},
	{"ucb", required_argument, NULL, ARG_UCB},
	
	{"bg_image_file", required_argument, NULL, ARG_BG_IMAGE_FILE},

	{"random", no_argument, NULL, ARG_RANDOM},
	{"selected", no_argument, NULL, ARG_SELECTED},
	{"random_pop", required_argument, NULL, ARG_RANDOM_POP},

	{"fish1", required_argument, NULL, ARG_FISH1},
	{"fish2", required_argument, NULL, ARG_FISH2},
	{"fish3", required_argument, NULL, ARG_FISH3},
	{"fish4", required_argument, NULL, ARG_FISH4},
	{"fish5", required_argument, NULL, ARG_FISH5},
	{"fish6", required_argument, NULL, ARG_FISH6},
	{"swordfish", required_argument, NULL, ARG_SWORDFISH},
	{"blowfish", required_argument, NULL, ARG_BLOWFISH},
	{"bdweller", required_argument, NULL, ARG_BDWELLER},
	{"fillmore", required_argument, NULL, ARG_FILLMORE},
	{"sherman", required_argument, NULL, ARG_SHERMAN},
	{"megan", required_argument, NULL, ARG_MEGAN},
	{"hunter", required_argument, NULL, ARG_HUNTER},
	{"prey", required_argument, NULL, ARG_PREY},
	{"lori", required_argument, NULL, ARG_LORI},
	{"ernest", required_argument, NULL, ARG_ERNEST},
	{"squid", required_argument, NULL, ARG_SQUID},
	{"hawthorne", required_argument, NULL, ARG_HAWTHORNE},

	{"eat", no_argument, NULL, ARG_EAT},
	{"explode", no_argument, NULL, ARG_EXPLODE},
	{"rebirth", no_argument, NULL, ARG_REBIRTH},
	{"scalediff", no_argument, NULL, ARG_SCALEDIFF},
	{"speed", required_argument, NULL, ARG_SPEED},
	{"scale", required_argument, NULL, ARG_SCALE},
	{"huntera", required_argument, NULL, ARG_HUNTERA},
	{"sworda", required_argument, NULL, ARG_SWORDA},
	{"window-id", required_argument, NULL, ARG_WINDOW_ID},
	{"root", no_argument, NULL, ARG_FULLSCREEN},
	{"bubble", required_argument, NULL, ARG_BUBBLE},
	{0,0,0,0}};

    b = bottom_get_settings_ptr();
    bub = bubble_get_settings_ptr();
    bg = background_get_settings_ptr();
    f = fish_get_settings_ptr();
    s = sound_get_settings_ptr();

    memcpy(f,&fish_settings, sizeof(Fish_settings));
    memcpy(bg,&background_settings, sizeof(Background_settings));
    memcpy(b,&bottom_settings, sizeof(Bottom_settings));
    memcpy(bub, &bubble_settings, sizeof(Bubble_settings));
    memcpy(s, &sound_settings, sizeof(Sound_settings));

    while ((c =
	    getopt_long_only(argc, argv, "", cmdline_options, NULL)) != -1){

	switch(c){

	case ARG_SEAFLOOR:
	    b->have_sea_floor = 1;
	    break;
	case ARG_SOUND:
	    s->on = 1;
	    break;
	case ARG_PLANTS:
	    b->max_plants = atoi(optarg);
	    break;
	case ARG_PLANTSCALE:
	    b->scale = atoi(optarg);
	    break;
	case ARG_BOTTOMANIMALS:
	    b->num_bottom_animals = atoi(optarg);
	    break;
	case ARG_OGG:
	    s->type_ogg = 1;
	    s->type_mp3 = 0;
	    break;
	case ARG_BG_SOLID:
	    bg->solid = 1;
	    bg->shaded = 0;
	    bg->wateralike = 0;
	    bg->image = 0;
	    break;
	case ARG_BG_SHADED:
	    bg->solid = 0;
	    bg->shaded = 1;
	    bg->wateralike = 0;
	    bg->image = 0;
	    break;
	case ARG_BG_WATERALIKE:
	    bg->solid = 0;
	    bg->shaded = 0;
	    bg->wateralike = 1;
	    bg->image = 0;
	    break;
	case ARG_BG_IMAGE:
	    bg->solid = 0;
	    bg->shaded = 0;
	    bg->wateralike = 0;
	    bg->image = 1;
	    break;
	case ARG_LCR:
	    bg->r1 = bg->r2 = atoi(optarg);
	    break;
	case ARG_LCG:
	    bg->g1 = bg->g2 = atoi(optarg);
	    break;
	case ARG_LCB:
	    bg->b1 = bg->b2 = atoi(optarg);
	    break;
	case ARG_UCR:
	    bg->r3 = atoi(optarg);
	    break;
	case ARG_UCG:
	    bg->g3 = atoi(optarg);
	    break;
	case ARG_UCB:
	    bg->b3 = atoi(optarg);
	    break;
	case ARG_BG_IMAGE_FILE:
	    bg->imagename = g_strdup_printf(optarg);
	    break;
	case ARG_RANDOM:
	    f->random_fish = 1;
	    f->selection_fish = 0;
	    break;
	case ARG_SELECTED:
	    f->random_fish = 0;
	    f->selection_fish = 1;
	    break;
	case ARG_RANDOM_POP:
	    f->num_fish = atoi(optarg);
	    break;
	case ARG_FISH1:
	    f->fish1 = atoi(optarg);
	    break;
	case ARG_FISH2:
	    f->fish2 = atoi(optarg);
	    break;
	case ARG_FISH3:
	    f->fish3 = atoi(optarg);
	    break;
	case ARG_FISH4:
	    f->fish4 = atoi(optarg);
	    break;
	case ARG_FISH5:
	    f->fish5 = atoi(optarg);
	    break;
	case ARG_FISH6:
	    f->fish6 = atoi(optarg);
	    break;
	case ARG_BLOWFISH:
	    f->blowfish = atoi(optarg);
	    break;
	case ARG_SWORDFISH:
	    f->swordfish = atoi(optarg);
	    break;
	case ARG_BDWELLER:
	    f->bdweller = atoi(optarg);
	    break;
	case ARG_FILLMORE:
	    f->fillmore = atoi(optarg);
	    break;
	case ARG_SHERMAN:
	    f->sherman = atoi(optarg);
	    break;
	case ARG_MEGAN:
	    f->megan = atoi(optarg);
	    break;
	case ARG_HUNTER:
	    f->hunter = atoi(optarg);
	    break;
	case ARG_PREY:
	    f->prey = atoi(optarg);
	    break;
	case ARG_LORI:
	    f->lori = atoi(optarg);
	    break;
	case ARG_ERNEST:
	    f->ernest = atoi(optarg);
	    break;
	case ARG_SQUID:
	    f->squid = atoi(optarg);
	    break;
	case ARG_HAWTHORNE:
	    f->hawthorne = atoi(optarg);
	    break;
	case ARG_EAT:
	    f->eat = 1;
	    break;
	case ARG_EXPLODE:
	    f->explode = 1;
	    break;
	case ARG_REBIRTH:
	    f->rebirth = 1;
	    break;
	case ARG_SCALEDIFF:
	    f->scale_diff = 1;
	    break;
	case ARG_SPEED:
	    f->speed = atoi(optarg);
	    break;
	case ARG_SCALE:
	    f->scale = atoi(optarg);
	case ARG_HUNTERA:
	    f->hunter_agr = atoi(optarg);
	    break;
	case ARG_SWORDA:
	    f->swordfish_agr = atoi(optarg);
	    break;
	case ARG_SOUNDPRG:
	    s->prg = g_strdup_printf(optarg);
	    break;
	case ARG_WINDOW_ID:
	    window_id = htoi(optarg);
	    break;
	case ARG_FULLSCREEN:
	    fullscreen = 1;
	    break;
	case ARG_BUBBLE:
	    bub->max_bubbles = atoi(optarg);
	    break;

     	default:
	    break;
	}
    }

    numfish += f->fish1;
    numfish += f->fish2;
    numfish += f->fish3;
    numfish += f->fish4;
    numfish += f->fish5;
    numfish += f->fish6;
    numfish += f->swordfish;
    numfish += f->blowfish;
    numfish += f->fillmore;
    numfish += f->sherman;
    numfish += f->prey;
    numfish += f->hunter;
    numfish += f->lori;
    numfish += f->ernest;
    numfish += f->squid;
    numfish += f->megan;
    numfish += f->bdweller;
    numfish += f->hawthorne;

    if(f->selection_fish)
	f->num_fish = numfish;

}

int main(int argc, char **argv)
{
    gtk_init(&argc, &argv);
    srand(time(NULL));
    screensaver_init_param(argc, argv);
    screensaver_init(0);
    screensaver_main();
    return 0;
}


AquariumData *aquarium_get_settings_ptr(void)
{
    return &ad;
}

unsigned char *aquarium_install_path(void)
{
    return applet_defines.image_path;
}


void aquarium_draw_image(int x, int y, int idx, int rev, SA_Image *image)
{
    screensaver_draw_image(x, y, idx, rev, image);
}

void aquarium_draw_pic_alpha(SA_Image *image, int w, int h, int x, int y, int idx, int alpha)
{
    screensaver_draw_image(x, y, idx, 0, image);
} 


void aquarium_clean_image(int x, int y, int w, int h)
{
    screensaver_clean(x, y, w, h);
}

GdkPixbuf *gai_load_image(char *fname)
{
    GError *msg = NULL;
    GdkPixbuf *pix;
    char *full_name;

    full_name = g_strdup_printf("%s/%s",applet_defines.image_path,fname);
    pix = gdk_pixbuf_new_from_file(full_name, &msg);
    if(!pix)
	printf("%s\n",msg->message);
    g_free(full_name);

    return pix;
    
}
void gai_display_error_continue(char *msg)
{
    printf("%s\n",msg);
}
